
#include "config.h"
#include "lcd-routines.h"
#include "resources.h"

#include <avr/io.h>
#include <util/delay.h>
#include <avr/sleep.h>
#include <stdlib.h>
#include <string.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>

struct Diode {
	uint8_t Anode;
	uint8_t Cathode;
	int Voltage;
};

void CheckPins(uint8_t HighPin, uint8_t LowPin, uint8_t TristatePin);
void DischargePin(uint8_t PinToDischarge, uint8_t DischargeDirection);
unsigned int ReadADC(uint8_t mux);
void GetGateThresholdVoltage(void);


#define R_DDR DDRB
#define R_PORT PORTB

//  Port fr die Testwiderstnde
//  Die Widerstnde mssen an die unteren 6 Pins des Ports angeschlossen werden,
//  und zwar in folgender Reihenfolge:
//  RLx = 680R-Widerstand fr Test-Pin x
//  RHx = 470k-Widerstand fr Test-Pin x
//
//  RL1 an Pin 0
//  RH1 an Pin 1
//  RL2 an Pin 2
//  RH2 an Pin 3
//  RL3 an Pin 4
//  RH3 an Pin 5


#define ADC_PORT PORTC
#define ADC_DDR DDRC
#define ADC_PIN PINC
#define TP1 PC0
#define TP2 PC1
#define TP3 PC2

//  Dieser Port muss ber einen ADC verfgen (beim Mega8 also PORTC).
//  Fr die Test-Pins mssen die unteren 3 Pins dieses Ports benutzt werden.
//  Bitte die Tefinitionen fr TP1, TP2 und TP3 nicht ndern!

#define ON_DDR DDRD
#define ON_PORT PORTD
#define ON_PIN_REG PIND
#define ON_PIN PD6	//Pin, der auf high gezogen werden muss, um Schaltung in Betrieb zu halten
#define RST_PIN PD7	//Pin, der auf low gezogen wird, wenn der Einschalt-Taster gedrckt wird

//Bauteile
#define PART_NONE 0
#define PART_DIODE 1
#define PART_TRANSISTOR 2
#define PART_FET 3
#define PART_TRIAC 4
#define PART_THYRISTOR 5
#define PART_RESISTOR 6
//Ende (Bauteile)
//Spezielle Definitionen fr Bauteile
#define PART_MODE_N_E_MOS 1
#define PART_MODE_P_E_MOS 2
#define PART_MODE_N_D_MOS 3
#define PART_MODE_P_D_MOS 4
#define PART_MODE_N_JFET 5
#define PART_MODE_P_JFET 6

#define PART_MODE_NPN 1
#define PART_MODE_PNP 2

struct Diode diodes[3];
uint8_t NumOfDiodes;

uint8_t b,c,e;		//Anschlsse des Transistors
unsigned long hfe;	//Verstrkungsfaktor
uint8_t PartReady;	//Bauteil fertig erkannt
unsigned int hfe1, hfe2;	//Verstrkungsfaktoren
unsigned int uBE1, uBE2;
uint8_t PartMode;
uint8_t ra, rb;			//Widerstands-Pins
unsigned int rv1, rv2;	//Spannungsabfall am Widerstand

uint8_t PartFound;	//das gefundene Bauteil
char outval[6];

int main(void) 
{
  //Einschalten
  ON_DDR |= _BV(ON_PIN);
  ON_PORT |= _BV(ON_PIN) | _BV(RST_PIN);	//Strom an und Pullup fr Reset-Pin

  //ADC-Init
  ADCSRA = _BV(ADEN) | _BV(ADPS1) | _BV(ADPS0);	//Vorteiler=8

  lcd_init();

  if (MCUCSR & _BV(WDRF)) 
  {
      // berprfen auf Watchdog-Reset 
      // Das tritt ein, wenn der Watchdog 2s nicht zurckgesetzt wurde
      // Kann vorkommen, wenn sich das Programm in einer Endlosschleife "verheddert" hat.

      // Timeout-Meldung
    lcd_WriteResource(resTestTimedOut);
    _delay_ms(3000);

#ifdef AUTOPOWEROFF
      // abschalten
    ON_PORT &= ~(1<<ON_PIN);

      // should never be reached - if reached - continue to next test
#else
      // wait until user presses switch
    while (ON_PIN_REG & _BV(RST_PIN)) 
      ;
#endif
  }

    // Custom-Zeichen
  lcd_DefineChar(LCD_CHAR_DIODE);
    // Diodensymbol in LCD laden
  lcd_WriteResource(resDiodeIcon);

    // Einsprungspunkt, wenn Start-Taste im Betrieb erneut gedrckt wird
  while (1)
  {
    wdt_enable(WDTO_2S);	//Watchdog an
    PartFound = PART_NONE;
    NumOfDiodes = 0;
    PartReady = 0;
    PartMode = 0;
  
    lcd_clear();

#ifdef CHECKSUPPLY
      // Versorgungsspannung messen
    ReadADC(5 | _BV(REFS1));	//Dummy-Readout
    hfe1 = ReadADC(5 | _BV(REFS1)); 	//mit interner Referenz

      // Vcc < 7,6V; Warnung anzeigen
    if (hfe1 < 650) 
    {
        // Anzeige: "Batterie"
      lcd_WriteResource(resBat);

        // Vcc < 7,15V; zuverlssiger Betrieb nicht mehr mglich
      if (hfe1 < 600)
      {
          // Batterie leer!
        lcd_WriteResource(resBatEmpty);
        _delay_ms(1000);

          // abschalten
        ON_PORT &= ~(1<<ON_PIN);

        return 0;
      }
        // Batterie schwach
      lcd_WriteResource(resBatWeak);
      lcd_Line2();
    }
#endif

      // Test beginnen

      // String: Test luft
    lcd_WriteResource(resTestRunning);	
    
      // Alle 6 Kombinationsmglichkeiten fr die 3 Pins prfen
    CheckPins(TP1, TP2, TP3);
    CheckPins(TP1, TP3, TP2);
    CheckPins(TP2, TP1, TP3);
    CheckPins(TP2, TP3, TP1);
    CheckPins(TP3, TP2, TP1);
    CheckPins(TP3, TP1, TP2);

      // Fertig, jetzt folgt die Auswertung
    GetGateThresholdVoltage();

    lcd_clear();

    if ((PartFound == PART_DIODE) && (NumOfDiodes == 1)) 
    {
        // Standard-Diode
      lcd_WriteResource(resDiode);	//"Diode: "
      lcd_string("A=");
      lcd_data(diodes[0].Anode + 49);
      lcd_string(";K=");
      lcd_data(diodes[0].Cathode + 49);
      lcd_Line2();	//2. Zeile
      lcd_WriteResource(resUf);	//"Uf = "
      lcd_string(itoa(diodes[0].Voltage, outval, 10));
      lcd_WriteResource(resmV);

      //goto end;
    } 
    else if ((PartFound == PART_DIODE) && (NumOfDiodes == 2)) 
    {
      //Doppeldiode
      if (diodes[0].Anode == diodes[1].Anode) 
      {
        //Common Anode
        lcd_WriteResource(resDualDiode);	//Doppeldiode
        lcd_string("CA");
        lcd_Line2(); //2. Zeile
        lcd_string("A=");
        lcd_data(diodes[0].Anode + 49);
        lcd_string(";K1=");
        lcd_data(diodes[0].Cathode + 49);
        lcd_string(";K2=");
        lcd_data(diodes[1].Cathode + 49);
      } 
      else if (diodes[0].Cathode == diodes[1].Cathode) 
      {
        //Common Cathode
        lcd_WriteResource(resDualDiode);	//Doppeldiode
        lcd_string("CC");
        lcd_Line2(); //2. Zeile
        lcd_string("K=");
        lcd_data(diodes[0].Cathode + 49);
        lcd_WriteResource(resA1);		//";A1="
        lcd_data(diodes[0].Anode + 49);
        lcd_WriteResource(resA2);		//";A2="
        lcd_data(diodes[1].Anode + 49);
      } 
      else if ( (diodes[0].Cathode == diodes[1].Anode) && 
                (diodes[1].Cathode == diodes[0].Anode) )
      {
        //Antiparallel
        lcd_WriteResource(resTwoDiodes);	//2 Dioden
        lcd_Line2(); //2. Zeile
        lcd_WriteResource(resAntiparallel);	//Antiparallel
      } 
      else if ( (diodes[0].Cathode == diodes[1].Anode) || 
                (diodes[1].Cathode == diodes[0].Anode) ) 
      {
        //Serienschaltung
        lcd_WriteResource(resTwoDiodes);	//2 Dioden
        lcd_Line2(); //2. Zeile
        lcd_WriteResource(resInSeries); //"in Serie A="

        if (diodes[0].Cathode == diodes[1].Anode) 
        {
          lcd_data(diodes[0].Anode + 49);
          lcd_string(";K=");
          lcd_data(diodes[1].Cathode + 49);
        } 
        else 
        {
          lcd_data(diodes[1].Anode + 49);
          lcd_string(";K=");
          lcd_data(diodes[0].Cathode + 49);
        }
      }

      //goto end;
    } 
      // Transistor
    else if (PartFound == PART_TRANSISTOR) 
    {
        // Wenn 2. Prfung nie gemacht, z.B. bei Transistor mit Schutzdiode
      if(PartReady == 0) 
      {
        hfe2 = hfe1;
        uBE2 = uBE1;
      }

        // Wenn der Verstrkungsfaktor beim ersten Test hher war: C und E vertauschen!
      if (hfe1 > hfe2)
      {
        hfe2 = hfe1;
        uBE2 = uBE1;

        uint8_t tmp = c;
        c = e;
        e = tmp;
      }

        // NPN oder PNP Ausgabe
      if (PartMode == PART_MODE_NPN)
        lcd_WriteResource(resNPN);
      else
        lcd_WriteResource(resPNP);

      lcd_WriteResource(resTransB);	//B=
      lcd_data(b + 49);
      lcd_WriteResource(resTransC);	//;C=
      lcd_data(c + 49);
      lcd_WriteResource(resTransE);	//;E=
      lcd_data(e + 49);
      lcd_Line2(); //2. Zeile

        // Verstrkungsfaktor berechnen

        // hFE = Emitterstrom / Basisstrom
      hfe = hfe2;
      hfe *= 691;	//Verhltnis von High- zu Low-Widerstand

      if (uBE2 < 11) 
        uBE2 = 11;

      hfe /= uBE2;
      hfe2 = (unsigned int) hfe;
      lcd_WriteResource(reshFE);	//"hFE = "
      lcd_string(utoa(hfe2, outval, 10));

        // Transistor mit Schutzdiode?
      if(NumOfDiodes == 3)
      {
        lcd_command(0xCD); //Cursor auf Zeile 2, Zeichen 13
        lcd_data(LCD_CHAR_DIODE);	//Diode anzeigen
      }

      //goto end;		
    } 
      // JFET oder MOSFET
    else if (PartFound == PART_FET) 
    {
      if (PartMode & 1)
        lcd_data('N');  // N-Kanal
      else
        lcd_data('P');	// P-Kanal

      if ((PartMode==PART_MODE_N_D_MOS) || (PartMode==PART_MODE_P_D_MOS)) 
      {
        lcd_WriteResource(resdmode);	//"-D"
        lcd_WriteResource(resMOSFET);	//"-MOS"
      } 
      else if ((PartMode==PART_MODE_N_JFET) || (PartMode==PART_MODE_P_JFET)) 
      {
        lcd_WriteResource(resJFET);	//"-JFET"
      }
      else 
      {
        lcd_WriteResource(resemode);	//"-E"
        lcd_WriteResource(resMOSFET);	//"-MOS"
      }

      lcd_Line2(); //2. Zeile
      lcd_WriteResource(resgds);	//"GDS="
      lcd_data(b + 49);
      lcd_data(c + 49);
      lcd_data(e + 49);

        // MOSFET mit Schutzdiode?
      if (NumOfDiodes == 1)
        lcd_data(LCD_CHAR_DIODE);	//Diode anzeigen
      else
        lcd_data(' ');	//Leerzeichen

        // Anreicherungs-MOSFET
      if (PartMode < 3) 
      {
        lcd_WriteResource(resvt);
        lcd_string(outval);	//Gate-Schwellspannung, wurde zuvor ermittelt
        lcd_data('m');
      }

      //goto end;
    } 
      // Thyristor
    else if (PartFound == PART_THYRISTOR) 
    {
      lcd_WriteResource(resThyristor);	//"Thyristor"
      lcd_Line2(); //2. Zeile
      lcd_string("G=");
      lcd_data(b + 49);
      lcd_string(";A=");
      lcd_data(c + 49);
      lcd_string(";K=");
      lcd_data(e + 49);

      //goto end;
    } 
      // Triac
    else if (PartFound == PART_TRIAC) 
    {
      lcd_WriteResource(resTriac);	//"Triac"
      lcd_Line2(); //2. Zeile
      lcd_string("G=");
      lcd_data(b + 49);
      lcd_WriteResource(resA1);		//";A1="
      lcd_data(c + 49);
      lcd_WriteResource(resA2);		//";A2="
      lcd_data(e + 49);

      //goto end;
    } 
    else if (PartFound == PART_RESISTOR) 
    {
      lcd_WriteResource(resResistor); //"Widerstand: "
      lcd_data(ra + 49);	//Pin-Angaben
      lcd_data('-');
      lcd_data(rb + 49);
      lcd_Line2(); //2. Zeile
      lcd_string ("R = ");

        // berprfen, wie weit die an den Testwiderstnden anliegenden Spannungen von 512 abweichen
      if (rv1 > 512) 
        hfe1 = (rv1 - 512);
      else
        hfe1 = (512 - rv1);

      if (rv2 > 512)
        hfe2 = (rv2 - 512);
      else
        hfe2 = (512 - rv2);

      if (hfe1 > hfe2)  
      {
        rv1 = rv2;	//Ergebnis verwenden, welches nher an 512 liegt (bessere Genauigkeit)
        rv2 = 4700;	//470k-Testwiderstand
      } 
      else
        rv2 = 680;	//680R-Testwiderstand

      if ( !rv1 ) 
        rv1 = 1;

      hfe = (long)((long)((long)rv2 * (long)rv1) / (long)(1023 - rv1));	//Widerstand berechnen
      hfe1 = (unsigned int)hfe;
      utoa(hfe1,outval,10);

        // 470k-Widerstand?
      if (rv2 == 4700) 
      {
          // Lnge ntig, um Komma anzuzeigen
        ra = strlen(outval);	

        for(rb=0; rb < ra; rb++)
        {
          lcd_data(outval[rb]);

          if ( rb == (ra-2))
            lcd_data('.');  // Komma
        }

        lcd_data ('k'); //Kilo-Ohm, falls 470k-Widerstand verwendet
      }
      else
        lcd_string(outval);

      lcd_data(LCD_CHAR_OMEGA);	//Omega fr Ohm

      //goto end;
    }
    else
    {
      if (NumOfDiodes == 0) 
      {
          // Keine Dioden gefunden
        lcd_WriteResource(resTestFailed1); //"Kein,unbek. oder"
        lcd_Line2(); //2. Zeile
        lcd_WriteResource(resTestFailed2); //"defektes "
        lcd_WriteResource(resBauteil);
      } 
      else 
      {
        lcd_WriteResource(resBauteil);
        lcd_WriteResource(resUnknown); //" unbek."
        lcd_Line2(); //2. Zeile
        lcd_WriteResource(resOrBroken); //"oder defekt"
        lcd_data(NumOfDiodes + 48);
        lcd_data('d');
      }
    }

      // wait for user to release switch
    while (!(ON_PIN_REG & (1<<RST_PIN)))
      ;

    _delay_ms(200);

#ifdef AUTOPOWEROFF
    for(hfe1 = 0;hfe1<10000;hfe1++)
    {
        // switch pressed?
      if (!(ON_PIN_REG & (1<<RST_PIN)))
      {
          // yes, so leave the loop and continue the outer
          // loop to the next test
        break;
      }

      wdt_reset();
      _delay_ms(1);
    }

      // if switch is still not pressed, switch off
    if ((ON_PIN_REG & _BV(RST_PIN)))
       ON_PORT &= ~_BV(ON_PIN);   
#endif

    wdt_disable();	//Watchdog aus

      // wait till the user wants to test again
    while((ON_PIN_REG & (1<<RST_PIN)))
      ;
  } // while (1) - main loop

    // never reached
  return 0;
}

  //  Funktion zum Testen der Eigenschaften des Bauteils bei der angegebenen Pin-Belegung
  //  Parameter:
  //  HighPin: Pin, der anfangs auf positives Potenzial gelegt wird
  //  LowPin: Pin, der anfangs auf negatives Potenzial gelegt wird
  //  TristatePin: Pin, der anfangs offen gelassen wird
  //
  //  Im Testverlauf wird TristatePin natrlich auch positiv oder negativ geschaltet.
void CheckPins(uint8_t HighPin, uint8_t LowPin, uint8_t TristatePin)
{
  unsigned int adcv[4];
  uint8_t tmpval, tmpval2;

    // HighPin wird fest auf Vcc gelegt
    // LowPin wird ber R_L auf GND gelegt
    // TristatePin wird hochohmig geschaltet, dafr ist keine Aktion ntig
  wdt_reset();

    // Pins setzen
  tmpval = (LowPin * 2);			//ntig wegen der Anordnung der Widerstnde
  R_DDR = (1<<tmpval);			//Low-Pin auf Ausgang und ber R_L auf Masse
  R_PORT = 0;
  ADC_DDR = (1<<HighPin);			//High-Pin auf Ausgang
  ADC_PORT = (1<<HighPin);		//High-Pin fest auf Vcc
  _delay_ms(5);

    // Bei manchen MOSFETs muss das Gate (TristatePin) zuerst entladen werden
    // N-Kanal:
  DischargePin(TristatePin,0);

    // Spannung am Low-Pin ermitteln
  adcv[0] = ReadADC(LowPin);

    // Sperrt das Bauteil jetzt?
  if ( !(adcv[0] < 20) )
  {
      // sonst: Entladen fr P-Kanal (Gate auf Plus)
    DischargePin(TristatePin,1);
      // Spannung am Low-Pin ermitteln
    adcv[0] = ReadADC(LowPin);
  }

    // Wenn das Bauteil keinen Durchgang zwischen HighPin und LowPin hat
  if(adcv[0] < 20)
  {
      // ntig wegen der Anordnung der Widerstnde
    tmpval = (TristatePin * 2);
      // Tristate-Pin ber R_L auf Masse
    R_DDR |= (1<<tmpval);

    _delay_ms(2);

      // Spannung messen
    adcv[0] = ReadADC(LowPin);
    if(adcv[0] > 700)
    {
        // Bauteil leitet => pnp-Transistor o..
        // Verstrkungsfaktor in beide Richtungen messen
        // if(PartReady==1) goto testend;
      R_DDR &= ~(1<<tmpval);		//Tristate-Pin (Basis) hochohmig
      tmpval++;
      R_DDR |= (1<<tmpval);		//Tristate-Pin (Basis) ber R_H auf Masse

      _delay_ms(10);
      adcv[0] = ReadADC(LowPin);		//Spannung am Low-Pin (vermuteter Kollektor) messen
      adcv[2] = ReadADC(TristatePin);	//Basisspannung messen
      R_DDR &= ~(1<<tmpval);		//Tristate-Pin (Basis) hochohmig
      //Prfen, ob Test schon mal gelaufen
      if ((PartFound == PART_TRANSISTOR) || (PartFound == PART_FET))
      {
        PartReady = 1;
        hfe2 = adcv[0];
        uBE2 = adcv[2];
      }
      else
      {
        hfe1 = adcv[0];
        uBE1 = adcv[2];
      }

      if(adcv[2] > 200)
      {
        if (PartFound != PART_THYRISTOR)
        {
            // PNP-Transistor gefunden (Basis wird "nach oben" gezogen)
          PartFound = PART_TRANSISTOR;
          PartMode = PART_MODE_PNP;
        }
      }
      else
      {
        if (PartFound != PART_THYRISTOR)
        {
            // P-Kanal-MOSFET gefunden (Basis/Gate wird NICHT "nach oben" gezogen)
          PartFound = PART_FET;
          PartMode = PART_MODE_P_E_MOS;
        }
      }

      if (PartFound != PART_THYRISTOR)
      {
        b = TristatePin;
        c = LowPin;
        e = HighPin;
      }
    }

    // Tristate (vermutete Basis) auf Plus, zum Test auf npn
    ADC_PORT = 0;					//Low-Pin fest auf Masse
    tmpval = (TristatePin * 2);		//ntig wegen der Anordnung der Widerstnde
    tmpval2 = (HighPin * 2);		//ntig wegen der Anordnung der Widerstnde
    R_DDR = (1<<tmpval) | (1<<tmpval2);			//High-Pin und Tristate-Pin auf Ausgang
    R_PORT = (1<<tmpval) | (1<<tmpval2);		//High-Pin und Tristate-Pin ber R_L auf Vcc
    ADC_DDR = (1<<LowPin);			//Low-Pin auf Ausgang
    _delay_ms(10);

    adcv[0] = ReadADC(HighPin);		//Spannung am High-Pin messen
    if(adcv[0] < 500)
    {
      if(PartReady==1) goto testend;
      //Bauteil leitet => npn-Transistor o..

      //Test auf Thyristor:
      //Gate entladen

      R_PORT &= ~(1<<tmpval);			//Tristate-Pin (Gate) ber R_L auf Masse
      _delay_ms(10);
      R_DDR &= ~(1<<tmpval);			//Tristate-Pin (Gate) hochohmig
      //Test auf Thyristor
      _delay_ms(5);
      adcv[1] = ReadADC(HighPin);		//Spannung am High-Pin (vermutete Anode) erneut messen
			
      R_PORT &= ~(1<<tmpval2);	//High-Pin (vermutete Anode) auf Masse
      _delay_ms(10);
      R_PORT |= (1<<tmpval2);	//High-Pin (vermutete Anode) wieder auf Plus
      _delay_ms(5);
      adcv[2] = ReadADC(HighPin);		//Spannung am High-Pin (vermutete Anode) erneut messen
      if((adcv[1] < 500) && (adcv[2] > 900)) {	//Nach Abschalten des Haltestroms muss der Thyristor sperren
        //war vor Abschaltung des Triggerstroms geschaltet und ist immer noch geschaltet obwohl Gate aus => Thyristor
        if(PartFound == PART_THYRISTOR) {
          PartFound = PART_TRIAC;	//Dieses Ergebnis schon einmal dagewesen => Triac (Doppelthyristor)
          PartReady = 1;
          goto saveresult;
        } else {
          PartFound = PART_THYRISTOR;
          goto saveresult;
        }
      }
      //Test auf Transistor oder MOSFET
      tmpval++;
      R_DDR |= (1<<tmpval);		//Tristate-Pin (Basis) auf Ausgang
      R_PORT |= (1<<tmpval);		//Tristate-Pin (Basis) ber R_H auf Plus
      _delay_ms(30);
      adcv[0] = ReadADC(HighPin);		//Spannung am High-Pin (vermuteter Kollektor) messen
      adcv[2] = ReadADC(TristatePin);	//Basisspannung messen
      R_PORT &= ~(1<<tmpval);			//Tristate-Pin (Basis) hochohmig
      R_DDR &= ~(1<<tmpval);			//Tristate-Pin (Basis) auf Eingang

      if((PartFound == PART_TRANSISTOR) || (PartFound == PART_FET)) {	//prfen, ob Test schon mal gelaufen
        PartReady = 1;
        hfe2 = 1023 - adcv[0];
        uBE2 = 1023 - adcv[2];
      } else {
        hfe1 = 1023 - adcv[0];
        uBE1 = 1023 - adcv[2];
      }
      if(adcv[2] < 600) {
        PartFound = PART_TRANSISTOR;	//NPN-Transistor gefunden (Basis wird "nach unten" gezogen)
        PartMode = PART_MODE_NPN;
      } else {
        PartFound = PART_FET;			//N-Kanal-MOSFET gefunden (Basis/Gate wird NICHT "nach unten" gezogen)
        PartMode = PART_MODE_N_E_MOS;
      }
      saveresult:
      b = TristatePin;
      c = HighPin;
      e = LowPin;
    }
    ADC_DDR = 0;
    ADC_PORT = 0;
    //Fertig
  } else {	//Durchgang
    //Test auf N-JFET oder selbstleitenden N-MOSFET
    R_DDR |= (2<<(TristatePin*2));	//Tristate-Pin (vermutetes Gate) ber R_H auf Masse
    _delay_ms(20);
    adcv[0] = ReadADC(LowPin);		//Spannung am vermuteten Source messen
    R_PORT |= (2<<(TristatePin*2));	//Tristate-Pin (vermutetes Gate) ber R_H auf Plus
    _delay_ms(20);
    adcv[1] = ReadADC(LowPin);		//Spannung am vermuteten Source erneut messen
    //Wenn es sich um einen selbstleitenden MOSFET oder JFET handelt, msste adcv[1] > adcv[0] sein
    if(adcv[1]>(adcv[0]+100)) {
      //Spannung am Gate messen, zur Unterscheidung zwischen MOSFET und JFET
      ADC_PORT = 0;
      ADC_DDR = (1<<LowPin);	//Low-Pin fest auf Masse
      tmpval = (HighPin * 2);		//ntig wegen der Anordnung der Widerstnde
      R_DDR |= (1<<tmpval);			//High-Pin auf Ausgang
      R_PORT |= (1<<tmpval);			//High-Pin ber R_L auf Vcc
      _delay_ms(20);
      adcv[2] = ReadADC(TristatePin);		//Spannung am vermuteten Gate messen
      if(adcv[2]>800) {	//MOSFET
        PartFound = PART_FET;			//N-Kanal-MOSFET
        PartMode = PART_MODE_N_D_MOS;	//Verarmungs-MOSFET
      } else {	//JFET (pn-bergang zwischen G und S leitet)
        PartFound = PART_FET;			//N-Kanal-JFET
        PartMode = PART_MODE_N_JFET;
      }
      PartReady = 1;
      b = TristatePin;
      c = HighPin;
      e = LowPin;
    }
    ADC_PORT = 0;

    //Test auf P-JFET oder selbstleitenden P-MOSFET
    ADC_DDR = (1<<LowPin);	//Low-Pin (vermuteter Drain) fest auf Masse, Tristate-Pin (vermutetes Gate) ist noch ber R_H auf Plus
    tmpval = (HighPin * 2);			//ntig wegen der Anordnung der Widerstnde
    R_DDR |= (1<<tmpval);			//High-Pin auf Ausgang
    R_PORT |= (1<<tmpval);			//High-Pin ber R_L auf Vcc
    _delay_ms(20);
    adcv[0] = ReadADC(HighPin);		//Spannung am vermuteten Source messen
    R_PORT &= ~(2<<(TristatePin*2));	//Tristate-Pin (vermutetes Gate) ber R_H auf Masse
    _delay_ms(20);
    adcv[1] = ReadADC(HighPin);		//Spannung am vermuteten Source erneut messen
    //Wenn es sich um einen selbstleitenden P-MOSFET oder P-JFET handelt, msste adcv[0] > adcv[1] sein
    if(adcv[0]>(adcv[1]+100)) {
      //Spannung am Gate messen, zur Unterscheidung zwischen MOSFET und JFET
      ADC_PORT = (1<<HighPin);	//High-Pin fest auf Plus
      ADC_DDR = (1<<HighPin);		//High-Pin auf Ausgang
      _delay_ms(20);
      adcv[2] = ReadADC(TristatePin);		//Spannung am vermuteten Gate messen
      if(adcv[2]<200) {	//MOSFET
        PartFound = PART_FET;			//P-Kanal-MOSFET
        PartMode = PART_MODE_P_D_MOS;	//Verarmungs-MOSFET
      } else {	//JFET (pn-bergang zwischen G und S leitet)
        PartFound = PART_FET;			//P-Kanal-JFET
        PartMode = PART_MODE_P_JFET;
      }
      PartReady = 1;
      b = TristatePin;
      c = LowPin;
      e = HighPin;
    }

    tmpval2 = (2<<(2*HighPin));	//R_H
    tmpval = (1<<(2*HighPin));	//R_L
    ADC_PORT = 0;
    //Test auf Diode
    ADC_DDR = (1<<LowPin);	//Low-Pin fest auf Masse, High-Pin ist noch ber R_L auf Vcc
    DischargePin(TristatePin,1);	//Entladen fr P-Kanal-MOSFET
    adcv[0] = ReadADC(HighPin);
    R_DDR = tmpval2;	//High-Pin ber R_H auf Plus
    R_PORT = tmpval2;
    adcv[2] = ReadADC(HighPin);
    R_DDR = tmpval;	//High-Pin ber R_L auf Plus
    R_PORT = tmpval;
    DischargePin(TristatePin,0);	//Entladen fr N-Kanal-MOSFET
    adcv[1] = ReadADC(HighPin);
    R_DDR = tmpval2;	//High-Pin ber R_H  auf Plus
    R_PORT = tmpval2;
    adcv[3] = ReadADC(HighPin);
    /*Ohne das Entladen kann es zu Falscherkennungen kommen, da das Gate eines MOSFETs noch geladen sein kann.
      Die zustzliche Messung mit dem "groen" Widerstand R_H wird durchgefhrt, um antiparallele Dioden von
      Widerstnden unterscheiden zu knnen.
      Eine Diode hat eine vom Durchlassstrom relativ unabhngige Durchlassspg.
      Bei einem Widerstand ndert sich der Spannungsabfall stark (linear) mit dem Strom.
    */
    if(adcv[0] > adcv[1]) {
      adcv[1] = adcv[0];	//der hhere Wert gewinnt
      adcv[3] = adcv[2];
    }
    if(PartFound == PART_NONE) PartFound = PART_DIODE;	//Diode nur angeben, wenn noch kein anderes Bauteil gefunden wurde. Sonst gbe es Probleme bei Transistoren mit Schutzdiode
    if(NumOfDiodes < 3) {
      diodes[NumOfDiodes].Anode = HighPin;
      diodes[NumOfDiodes].Cathode = LowPin;
      diodes[NumOfDiodes].Voltage = (adcv[1]*60/13);	// ca. mit 4,6 multiplizieren, um aus dem ADC-Wert die Spannung in Millivolt zu erhalten
      NumOfDiodes++;
      for(uint8_t i=0;i<NumOfDiodes;i++) {
        if((diodes[i].Anode == LowPin) && (diodes[i].Cathode == HighPin)) {	//zwei antiparallele Dioden: Defekt oder Duo-LED
          if((adcv[3]*2) < adcv[1]) {	//Durchlassspannung fllt bei geringerem Teststrom stark ab => Defekt
            if(i<NumOfDiodes) {
              for(uint8_t j=i;j<(NumOfDiodes-1);j++) {
                diodes[j].Anode = diodes[j+1].Anode;
                diodes[j].Cathode = diodes[j+1].Cathode;
                diodes[j].Voltage = diodes[j+1].Voltage;
              }
            }
            NumOfDiodes -= 2;
          }
        }
      }
    }
  }
  tmpval2 = (2<<(2*HighPin));	//R_H
  tmpval = (1<<(2*HighPin));	//R_L
  ADC_PORT = 0;
  ///Test auf Widerstand
  ADC_DDR = (1<<LowPin);	//Low-Pin fest auf Masse
  R_DDR = tmpval;	//High-Pin ber R_L auf Plus
  R_PORT = tmpval;
  adcv[0] = ReadADC(HighPin);
  R_DDR = tmpval2;	//High-Pin ber R_H auf Plus
  R_PORT = tmpval2;
  adcv[1] = ReadADC(HighPin);
  if(((adcv[1] + 200) < adcv[0]) && (((adcv[1] * 3) / 2) < adcv[0]) && (adcv[0] > 45)) {	//Durchlassspannung fllt bei geringerem Teststrom stark ab und es besteht kein "Beinahe-Kurzschluss" => Widerstand
    if((PartFound == PART_DIODE) || (PartFound == PART_NONE)) PartFound = PART_RESISTOR;
    ra = HighPin;
    rb = LowPin;
    rv1 = adcv[0];
    rv2 = adcv[1];
  }
testend:
  ADC_DDR = 0;
  ADC_PORT = 0;
  R_DDR = 0;
  R_PORT = 0;
}

  // ADC-Wert des angegebenen Kanals auslesen und als unsigned int zurckgegen
unsigned int ReadADC(uint8_t mux)
{
  unsigned int adcx = 0;

  ADMUX = mux | _BV(REFS0);

    // 20 Messungen; fr bessere Genauigkeit
  for (uint8_t j=0; j < 20; j++)
  {
    ADCSRA |= (1<<ADSC);
    
    loop_until_bit_is_clear(ADCSRA, ADSC);

    adcx += ADCW;
  }
  adcx /= 20;
  return adcx;
}

void GetGateThresholdVoltage(void)
{
  uint8_t tmpval, tmpval2;
  unsigned int tmpint = 0;
  tmpval = (1<<(2*c) | (2<<(2*b)));
  tmpval2=(1<<(2*c));
  R_DDR = tmpval;        // Drain ber R_L auf Ausgang, Gate ber R_H auf Ausgang
  ADC_DDR=(1<<e)|(1<<b);	//Gate und Source auf Ausgang
  if((PartFound==PART_FET) && (PartMode == PART_MODE_N_E_MOS)) {
    ADC_PORT = 0;			//Gate und Source fest auf Masse
    R_PORT = tmpval;  	   // Drain ber R_L auf Plus, Gate ber R_H auf Plus
    _delay_ms(5);

    tmpval=(1<<c);
    ADC_DDR=(1<<e);          // Gate ber R_H langsam laden
    while (ADC_PIN&tmpval) {  // Warten, bis der MOSFET schaltet und Drain auf low geht
      wdt_reset();
      tmpint++;
      if(tmpint==0) break;	//Timeout fr Gate-Schwellspannungs-Messung
    }
    R_PORT=tmpval2;          // Gate hochohmig schalten
    R_DDR=tmpval2;          // Gate hochohmig schalten
  }
  else if ((PartFound==PART_FET) && (PartMode == PART_MODE_P_E_MOS))
  {
    ADC_PORT = (1<<e)|(1<<b);	//Gate und Source fest auf Plus
    R_PORT = 0;					//Drain ber R_L auf Masse, Gate ber R_H auf Masse
    _delay_ms(5);

    tmpval=(1<<c);
    ADC_DDR=(1<<e);           // Gate ber R_H langsam laden (Gate auf Eingang)
    ADC_PORT=(1<<e);          // Gate ber R_H langsam laden (Gate-Pullup aus)
    while (!(ADC_PIN&tmpval)) // Warten, bis der MOSFET schaltet und Drain auf high geht
    {
      wdt_reset();
      tmpint++;

        // Timeout fr Gate-Schwellspannungs-Messung
      if (!tmpint)
        break;
    }
    R_DDR=tmpval2;          // Gate hochohmig schalten
  }

  if(tmpint > 0)
  {
    tmpint = ReadADC(b);

    if (PartMode == PART_MODE_P_E_MOS)
      tmpint = 1023-tmpint;

    tmpint = (tmpint*39/8);
  }
  utoa(tmpint, outval, 10);
  
  R_DDR = 0;
  R_PORT = 0;
  ADC_DDR = 0;
  ADC_PORT = 0;
}

void DischargePin(uint8_t PinToDischarge, uint8_t DischargeDirection)
{
    // Anschluss eines Bauelementes kurz(10ms) auf ein bestimmtes Potenzial legen
    // Diese Funktion ist zum Entladen von MOSFET-Gates vorgesehen, um Schutzdioden u.. in MOSFETs erkennen zu knnen
    // Parameter:
    // PinToDischarge: zu entladender Pin
    // DischargeDirection: 0 = gegen Masse (N-Kanal-FET), 1= gegen Plus(P-Kanal-FET)
  uint8_t tmpval;
  
    // ntig wegen der Anordnung der Widerstnde
  tmpval = (PinToDischarge * 2);

    // R_L aus 
  if (DischargeDirection)
    R_PORT |= (1<<tmpval);

    // Pin auf Ausgang und ber R_L auf Masse
  R_DDR |= (1<<tmpval);

  _delay_ms(10);

    // Pin wieder auf Eingang
  R_DDR &= ~(1<<tmpval);

    // R_L aus
  if (DischargeDirection)
    R_PORT &= ~(1<<tmpval);
}















